/*
 * DistPrefixSum 
 * DistPrefixSum.x10 
 * Dave Hudak
 * Ohio Supercomputer Center
 */

public class DistPrefixSum {
	
	val prefixSumArray:  DistArray[Int](1);
	val prefixSumStrings:  DistArray[String](1);
	val localSums:  DistArray[Int](1);
	
	public def this() {
		prefixSumStrings = DistArray.make[String](Dist.makeUnique());
		localSums = DistArray.make[Int](Dist.makeUnique(), 0);
		prefixSumArray = DistArray.make[Int](Dist.makeUnique(), ([i]:Point)=>i);
	}
	
	public def this(length:Int) {
		prefixSumStrings = DistArray.make[String](Dist.makeUnique());
		localSums = DistArray.make[Int](Dist.makeUnique(), (Point)=>0);
		val prefixSumArrayDist = Dist.makeBlock(1..length, 0);
		prefixSumArray = DistArray.make[Int](prefixSumArrayDist, ([i]:Point)=>i);
		
		/*
		 * prefixSumArray = DistArray.make[Int](prefixSumArrayDist, (Point)=>0);
		 * for (p:Place in prefixSumArray.dist.places()) {
		 * at (p) {
		 * for ([i] in prefixSumArray | here) {
		 * prefixSumArray(i) = i;
		 * }
		 * }
		 * }
		 */
	}
	
	public def this(lowerBound:Int, upperBound:Int) {
		prefixSumStrings = DistArray.make[String](Dist.makeUnique());
		localSums = DistArray.make[Int](Dist.makeUnique(), (Point)=>0);
		val length = (upperBound - lowerBound) + 1;
		val prefixSumArrayDist = Dist.makeBlock(1..length, 0);
		prefixSumArray = DistArray.make[Int](prefixSumArrayDist,
				([i]:Point)=>lowerBound + i - 1);
		
		/*
		 * prefixSumArray = DistArray.make[Int](prefixSumArrayDist, (Point)=>0);
		 * for (p in prefixSumArray.dist.places()) {
		 * at (p) {
		 * for ([i] in prefixSumArray | here) {
		 * prefixSumArray(i) = lowerBound + (i - 1);
		 * }
		 * }
		 * }
		 */
	}
	
	public def str():String 
	{
		var s : String = "";								
		for (p in prefixSumArray.dist.places()) {
			finish {
				at (p) {
					var str : String = "";
					var first : Boolean = true;	
					for (pt in prefixSumArray | here) {
						if (first) {
							first = false;
						}
						else {
							str += ", ";
						}
						str += prefixSumArray(pt).toString();
					}
					prefixSumStrings(here.id) = str;
				}
			}
			s += (at(p) prefixSumStrings(here.id));
			if (p.id != (Place.MAX_PLACES - 1)) {
				s += ", ";
			} //if
		} //for p 
		return s;
	}
	
	
	public def lastElement():Int 
	{
		val lastIndex = prefixSumArray.region.size();
		val p = prefixSumArray.dist(lastIndex);
		return (at (p) prefixSumArray(lastIndex));
	}
	
	
	public def computeSum()
	{
		//Compute local prefix sums at each place
		//Also, save the total for each place
		finish {
			for (p in prefixSumArray.dist.places()) {
				async at (p) {
					localSums(here.id) = 0;
					var first : Boolean = true;							
					for ([i] in prefixSumArray|here) {
						localSums(here.id) += prefixSumArray(i);
						if (first) {
							first = false;
						}
						else {
							prefixSumArray(i) = prefixSumArray(i) + prefixSumArray(i-1);
						}
					}  //for i
				} //at
			} //for p
		} //finish
		finish {
			for (p in prefixSumArray.dist.places()) {
				async at (p) {
					//Compute the total needed to update each place 
					val placeId = here.id;
					var globalUpdate: Int = 0;
					for (var j:Int=0;j<placeId;j++) {
						val valj = j;
						globalUpdate += (at (Place.places()(valj)) localSums(here.id));
					}
					//Update prefix sums at each place with totals
					for ([i] in prefixSumArray.dist|here) {
						prefixSumArray(i) += globalUpdate;
					}  //for i
				} //at
			} //for p
		} //finish
	} //computeSum
}

