/*
 * MLPrefixSum 
 * MLPrefixSum.x10 
 * Dave Hudak
 * Ohio Supercomputer Center
 */

public class MLPrefixSum {
	
	val prefixSumStrings:  DistArray[String](1);
	val localSums:  DistArray[Int](1);
	val globalDist : Dist;
	val prefixSumArray:  DistArray[PrefixSum](1);
	
	public def this() {
		prefixSumStrings = DistArray.make[String](Dist.makeUnique());
		localSums = DistArray.make[Int](Dist.makeUnique(), 0);
		globalDist = Dist.makeUnique();
		val factory = (i:Point) => {
			val psFactory = new PrefixSumFactory();
			psFactory.getPrefixSum()
		};
		prefixSumArray = DistArray.make[PrefixSum](Dist.makeUnique(), factory);
	}
	
	public def this(psFactory:PrefixSumFactory) {
		prefixSumStrings = DistArray.make[String](Dist.makeUnique());
		localSums = DistArray.make[Int](Dist.makeUnique(), 0);
		globalDist = Dist.makeUnique();
		prefixSumArray = DistArray.make[PrefixSum](Dist.makeUnique(), psFactory.getPrefixSum());
	}
	
	public def this(psFactory:PrefixSumFactory, length:Int) {
		prefixSumStrings = DistArray.make[String](Dist.makeUnique());
		localSums = DistArray.make[Int](Dist.makeUnique(), (Point)=>0);
		globalDist = Dist.makeBlock(1..length, 0);
		val factory = (i:Point) => {
			val localDist = Dist.makeBlock(1..length, 0)|here;
			val localRegion = localDist.region;
			val localMin = localRegion.min(0);
			val localMax = localRegion.max(0);
			psFactory.getPrefixSum(localMin,localMax)
		};
		prefixSumArray = DistArray.make[PrefixSum](Dist.makeUnique(), factory);
	}
	
	public def str():String 
	{
		finish for (p in prefixSumArray.dist.places()) {
			async at (p) {
				val psObject = prefixSumArray(here.id);
				prefixSumStrings(here.id) = psObject.str();
			}  //async
		} //finish for
		var s : String = "";								
		for (p in prefixSumArray.dist.places()) {
			s += (at(p) prefixSumStrings(here.id));
			if (p.id != (Place.MAX_PLACES - 1)) {
				s += ", ";
			} //if
		} //for p 
		return s;
	}
	
	public def computeSum()
	{
		//Compute local prefix sums at each place
		//Also, save the total for each place
		finish {
			for (p in prefixSumArray.dist.places()) {
				async at (p) {
					val psObject = prefixSumArray(here.id);
					psObject.computeSum();
					localSums(here.id) = psObject.lastElement();
				} //async
			} //for p
		} //finish
		finish {
			for (p in prefixSumArray.dist.places()) {
				async at (p) {
					//Compute the total needed to update each place 
					val placeId = here.id;
					var globalUpdate: Int = 0;
					for (var j:Int=0;j<placeId;j++) {
						val valj = j;
						globalUpdate += localSums(here.id);
					}
					val psObject = prefixSumArray(here.id);
					val update = globalUpdate;
					psObject.incrementArray(update);          
				} //async
			} //for p
		} //finish
	} //computeSum
}