/*
 * AsyncPrefixSum 
 * AsyncPrefixSum.x10 
 * Dave Hudak
 * Ohio Supercomputer Center
 */
public class AsyncPrefixSum extends PrefixSum {
	
	public def this(length:Int) {
		super(length);
	}
	
	public def this(lowerBound:Int, upperBound:Int) {
		super(lowerBound, upperBound);
	}
	
	public def incrementArray(amount:Int)
	{
		val numActivities = 4;
		val arrayLength = prefixSumArray.size;
		val chunkSize = arrayLength/numActivities as Int;
		
		finish {
			for (activityId in 1..numActivities) {
				async {
					val low = 1 + (activityId - 1)*chunkSize;
					var hi:Int = (low + chunkSize) - 1;
					if (activityId == numActivities) {
						hi = arrayLength;
					} //if
					for ([i] in low..hi) {
						prefixSumArray(i) = prefixSumArray(i) + amount;
					} //for
				} //async
			} //for
		} //finish
	}
	
	public def computeSum()
	{
		val chunkSize = 4;
		val tempArray = new Array[Int](1..prefixSumArray.size, 0);
		val arrayLength = prefixSumArray.size;
		val logLength = Math.log(arrayLength as Double);
		val lnLength = logLength/Math.log(2.0);
		val clnLength = Math.ceil(lnLength);
		val numSteps = clnLength as Int;
		for ([stepNumber] in 1..numSteps) {
			val stepWidth = Math.pow(2, (stepNumber - 1)) as Int;
			val globalLow = stepWidth + 1;
			val globalHi = arrayLength;
			val aSize = (globalHi - globalLow + 1) as Double;
			val numChunks = aSize / (chunkSize as Double);
			val numActivities = Math.ceil(numChunks) as Int;
			Console.OUT.println("numActivities = "+numActivities);
			finish {
				for ([activityId] in 1..numActivities) {
					async {
						val low = globalLow + (activityId - 1)*chunkSize;
						var hi:Int = (low + chunkSize) - 1;
						if (activityId == numActivities) {
							hi = globalHi;
						}
						Console.OUT.println(
								"stepNumber = " + stepNumber + 
								                ", stepWidth = " + stepWidth + 
								                                 ", globalLow = " + globalLow +
								                                                  ", globalHi = " + globalHi +
								                                                                  ", numActivities = " + numActivities + 
								                                                                                       ", activityId = " + activityId +
								                                                                                                         ", low = " + low +
								                                                                                                                    ", hi = " + hi);
						for ([j] in low..hi) {
							tempArray(j) = prefixSumArray(j) + prefixSumArray(j-stepWidth);
						} //for j
					} //async
				} //for activityId
			} //finish
			finish {
				for ([activityId] in 1..numActivities) {
					async {
						val low = globalLow + (activityId - 1)*chunkSize;
						var hi:Int = (low + chunkSize) - 1;
						if (activityId == numActivities) {
							hi = globalHi;
						}
						for ([j] in low..hi) {
							prefixSumArray(j) = tempArray(j);
						} //for j
					}
				} //for activityId
			} //finish 
		} //for stepNumber  
	} //computeSum
	
}